"""
 http binding interface
 $Id: httpb.py,v 1.42 2006/02/23 21:10:10 tofu Exp $
"""
from twisted.python import components
from twisted.web import server, resource
from twisted.internet import defer, task
from twisted.python import log
from twisted.application import service

from zope.interface import Interface, implements

from xmpp import component
use_xmlstream = False
try:
    from twisted.words.xish import domish
    from twisted.words.protocols.jabber import xmlstream
    use_xmlstream = True
except:
    from twisted.xish import domish

import sha, time
import session, error

import punjab
from xmpp import client, ns


NS_BIND = 'http://jabber.org/protocol/httpbind'
NS_FEATURES = 'http://etherx.jabber.org/streams'


# make httpb body class, similar to xmlrpclib
# 
class HttpbParse:
    """
    An xml parser for parsing the body elements.
    """
    def __init__(self):
        """
        Call reset to initialize object
        """
        self._reset()

    def parse(self, buf):
        """
        Parse incoming xml and return the body and its children in a list
        """
        self.stream.parse(buf)
        # return the doc element and its children in a list
        return self.body, self.xmpp_elements 
    
    def serialize(self, obj):
        """
        Turn object into a string type
        """
        if isinstance(obj, domish.Element):
            obj = obj.toXml()
        return obj

    def onDocumentStart(self, rootelem):
        """
        The body document has started.
        
        This should be a body.
        """
        if rootelem.name == 'body':
            self.body = rootelem

    def onElement(self, element):
        """
        A child element has been found.
        """
        if isinstance(element, domish.Element):
            
            self.xmpp_elements.append(element)
        else:
            pass

    def _reset(self):
        """
        Setup the parser
        """
        self.stream = domish.elementStream()
        self.stream.DocumentStartEvent = self.onDocumentStart
        self.stream.ElementEvent = self.onElement
        self.stream.DocumentEndEvent = self.onDocumentEnd
        self.body = ""
        self.xmpp_elements = []
        

    def onDocumentEnd(self):
        """
        Body End
        """
        pass

class IHttpbService(Interface):
    """
    Interface for http binding class
    """
    def __init__(self, verbose):
        """ """

    def startSession(self, body):
        """ Start a punjab jabber session """

    def endSession(self, session):
        """ end a punjab jabber session """

    def onExpire(self, session_id):
        """ preform actions based on when the jabber connection expires """

    def parseBody(self, body):
        """ parse a body element """


    def error(self, error):
        """ send a body error element """


    def inSession(self, body):
        """ """

    def getXmppElements(self, body, session):
        """ """
        

        
class IHttpbFactory(Interface):
    """
    Factory class for generating binding sessions.
    """
    def startSession(self):
        """ Start a punjab jabber session """

    def endSession(self, session):
        """ end a punjab jabber session """

    def parseBody(self, body):
        """ parse an body element """

    def buildProtocol(self, addr):
        """Return a protocol """

        
    
class Httpb(resource.Resource):
    """
    Http resource to handle BOSH requests.
    """
    isLeaf = True
    def __init__(self, service, v = 0):
        """Initialize.
        """
        resource.Resource.__init__(self)
        self.service  = service
        self.hp       = None
        self.children = {}
        self.client   = 0
        self.verbose  = v
        self.polling = self.service.polling or 15
                
    def render_GET(self, request):
        """
        GET is not used, print docs.
        """
        return """<html>
                 <body>
                 <a href='http://www.xmpp.org/extensions/xep-0124.html'>XEP-0124</a> - BOSH
                 </body>
               </html>"""

    def render_POST(self, request):
        """
        Parse received xml
        """
        request.content.seek(0, 0)
        if self.service.v:
            log.msg('HEADERS %s:' % (str(time.time()),))
            log.msg(request.received_headers)
            log.msg("HTTPB POST : ")
            log.msg(str(request.content.read()))
            request.content.seek(0, 0)       

        self.hp       = HttpbParse()
        try:
            body_tag, xmpp_elements = self.hp.parse(request.content.read()) 
            self.hp._reset()

            if body_tag.name != "body":
                self.send_http_error(400, request)
                return server.NOT_DONE_YET            
        except:
            log.err()
            # reset parser, just in case
            self.hp._reset()
            self.send_http_error(400, request) 
            return server.NOT_DONE_YET
        else:
            if self.service.inSession(body_tag):
                # sid is an existing session
                if body_tag.getAttribute('rid'):
                    request.rid = body_tag['rid']
                    log.msg(request.rid)
                
                s, d = self.service.parseBody(body_tag, xmpp_elements)
                d.addCallback(self.return_httpb, s, request)
            elif body_tag.hasAttribute('sid'):
                log.msg("no sid is found but the body element has a 'sid' attribute")
                # This is an error, no sid is found but the body element has a 'sid' attribute
                self.send_http_error(404, request)
                return server.NOT_DONE_YET
            else:
                # start session
                s, d = self.service.startSession(body_tag, xmpp_elements)
                d.addCallback(self.return_session, s, request)
                
            # Add an error back for returned errors
            d.addErrback(self.return_error, request)
        return server.NOT_DONE_YET
        

    def return_session(self, data, session, request):
        # create body
        if session.xmlstream is None:
            self.send_http_error(200, request, 'remote-connection-failed',
                                 'terminate')
            return server.NOT_DONE_YET
        
        b = domish.Element((NS_BIND, "body"))       
        # if we don't have an authid, we have to fail
        if session.authid != 0:
            b['authid'] = session.authid
        else:
            self.send_http_error(500, request, 'internal-server-error',
                                 'terminate')
            return server.NOT_DONE_YET
        
        b['sid']  = session.sid
        b['wait'] = str(session.wait)
        if session.secure == 0:
            b['secure'] = 'false'
        else:
            b['secure'] = 'true'
            
        b['inactivity'] = str(session.inactivity)
        ##b['polling'] = '15' # TODO: make this configurable
        b['polling'] = str(self.polling) 
        b['requests'] = str(session.hold + 1)
        b['window'] = str(session.window)
        
        punjab.uriCheck(b, NS_BIND)
        if session.attrs.has_key('content'):
            b['content'] = session.attrs['content']

        # We need to send features
        while len(data) > 0:
            felem = data.pop(0)
            b.addChild(felem)
        
        self.return_body(request, b)

    def return_httpb(self, data, session, request):
        # create body                
        b = domish.Element((NS_BIND, "body"))
        punjab.uriCheck(b, NS_BIND)
        session.touch()
        if data:
            b.children += data
        
        self.return_body(request, b, session.charset)        

    
    def return_error(self, e, request):
        echildren = []
        
        try:
            # TODO - clean this up and make errors better
            if getattr(e.value,'stanza_error',None):
                ec = getattr(e.value, 'children', None)
                if ec:
                    echildren = ec
                    
                self.send_http_error(error.conditions[str(e.value.stanza_error)]['code'],
                                     request,
                                     condition = str(e.value.stanza_error),
                                     typ = error.conditions[str(e.value.stanza_error)]['type'],
                                     children=echildren)
                
                return  server.NOT_DONE_YET
            elif e.value:
                self.send_http_error(error.conditions[str(e.value)]['code'],
                                     request,
                                     str(e.value),
                                     error.conditions[str(e.value)]['type'])
                return  server.NOT_DONE_YET
            else:
                self.send_http_error(500, request, 'internal-server-error', 'error', e)
        except:
            log.err()
            pass

    
    def return_body(self, request, b, charset="utf-8"):
        request.setResponseCode(200)
        bxml = b.toXml(prefixes=ns.XMPP_PREFIXES).encode(charset,'replace')
        request.setHeader('content-type', 'text/xml')
        request.setHeader('content-length', len(bxml))
        if self.service.v:
            log.msg('\n\nRETURN HTTPB %s:' % (str(time.time()),))
            log.msg(bxml)
        request.write(bxml)
        if getattr(request, 'rid', None):
            log.msg(request.rid)
        
        request.finish()
            
    def send_http_error(self, code, request, condition = 'undefined-condition', typ = 'terminate', data = '', charset = 'utf-8', children=None):
        request.setResponseCode(int(code))
        
        b = domish.Element((NS_BIND, "body"))
        if condition:
            b['condition'] = str(condition)
        else:
            b['condition'] = 'undefined-condition'

        if typ:
            b['type']      = str(typ)
        else:
            b['type']      = 'terminate'
        punjab.uriCheck(b, NS_BIND)
        bxml           = b.toXml().encode(charset, 'replace')

        if children:
            b.children += children
        
        log.msg('HTTPB Error %d' %(int(code),))
        
        if int(code) != 400 and int(code) != 404 and int(code) != 403:
            if data != '':
                if condition == 'see-other-uri':
                    b.addElement('uri', None, content = str(data))
                else:
                    t = b.addElement('text', content = str(data))
                    t['xmlns'] = 'urn:ietf:params:xml:ns:xmpp-streams'
                    
            bxml           = b.toXml().encode(charset, 'replace')
            log.msg('HTTPB Return Error: ' + str(code) + ' -> ' + bxml)
            request.setHeader("content-type", "text/xml")
            request.setHeader("content-length", len(bxml))    
            request.write(bxml)
        else:
            request.setHeader("content-length", "0")
        request.finish()


components.registerAdapter(Httpb, IHttpbService, resource.IResource)


class HttpbService(punjab.Service):

    implements(IHttpbService)

    def __init__(self, verbose = 0, polling = 15):
        self.v  = verbose
        self.sessions = {}
        self.counter  = 0
        self.polling = polling
        # self.expired  = {}

        # run a looping call to do pollTimeouts on sessions
        self.poll_timeouts = task.LoopingCall(self._doPollTimeOuts)

        self.poll_timeouts.start(3) # run every 3 seconds

    def _doPollTimeOuts(self):
        """
        Call poll time outs on sessions that have waited too long.
        """
        time_now = time.time() + 2.9 # need a number to offset the poll timeouts
        for session in self.sessions.itervalues():
            if len(session.waiting_requests)>0:
                for wr in session.waiting_requests:
                    if time_now - wr.wait_start >= wr.timeout:
                        wr.delayedcall(wr.deferred)

            
    def startSession(self, body, xmpp_elements):
        """ Start a punjab jabber session """
    
        # look for rid
        if not body.hasAttribute('rid') or body['rid']=='':
            log.msg('start session called but we had a rid')
            return None, defer.fail(error.NotFound)
                
        # look for to
        if not body.hasAttribute('to') or body['to']=='':
            return None, defer.fail(error.BadRequest)
        
        # look for wait
        if not body.hasAttribute('wait') or body['wait']=='':
            body['wait'] = 3
                
        # look for lang
        lang = None
        if not body.hasAttribute("xml:lang") or body['xml:lang']=='':
            for k in body.attributes:
                if isinstance(k, tuple):
                    if str(k[1])=='lang' and body.getAttribute(k) !='':
                        lang = body.getAttribute(k)
        if lang:
            body['lang'] = lang
        if not body.hasAttribute('inactivity'):
            body['inactivity'] = 60 
        return session.make_session(self, body.attributes)
                

    def parseBody(self, body, xmpp_elements):
        try:
            # grab session                    
            if body.hasAttribute('sid'):
                sid = str(body['sid'])
            else:
                log.msg('Session ID not found')
                return None, defer.fail(error.NotFound)

            if self.inSession(body):
                s = self.sessions[sid]
                s.touch() # any connection should be a renew on wait
            else:
                log.msg('session does not exist?')
                return None, defer.fail(error.NotFound)
            ##  XXX this seems to break xmpp:restart='true'  --vargas
            ##  (cf. http://www.xmpp.org/extensions/xep-0206.html#preconditions-sasl [Example 10])
##            if body.hasAttribute('to') and body['to']!='':
##                return s, defer.fail(error.BadRequest)
            
            # check for keys
            # TODO - clean this up
            foundNewKey = False
            
            if body.hasAttribute('newkey'):
                newkey = body['newkey']
                s.key = newkey
                foundNewKey = True
            try:
                if body.hasAttribute('key') and not foundNewKey:
                    if s.key is not None:
                        nk = sha.new(body['key'])
                        key = nk.hexdigest()
                        next_key = body['key']
                        if key == s.key:
                            s.key = next_key
                        else:
                            log.msg('Error in key')
                            return s, defer.fail(error.NotFound)                        
                    else:
                        log.err()
                        raise s, defer.fail(error.NotFound)
                        
            except:
                log.msg('HTTPB ERROR: ')
                log.err()
                return s, defer.fail(error.NotFound)
                    
            # need to check if this is a valid rid (within tolerance)
            if body.hasAttribute('rid') and body['rid']!='': 
                if s.cache_data.has_key(int(body['rid'])):
                    # implements issue 32 and returns the data returned on a dropped connection
                    return s, defer.succeed(s.cache_data[int(body['rid'])])
                if abs(int(body['rid']) - int(s.rid)) > s.window:
                    log.msg('This rid is invalid')
                    return  s, defer.fail(error.NotFound)
            else:
                log.msg('There is no rid on this request')
                return  s, defer.fail(error.NotFound)
                  
            return s, self._parse(s, body, xmpp_elements)
            
            
        except:
            log.err()
            return  s, defer.fail(error.InternalServerError)

            
    def onExpire(self, session_id):
        """ preform actions based on when the jabber connection expires """
        log.msg('expire (%s)' % (str(session_id),))
        log.msg(len(self.sessions.keys()))
        
    def _parse(self, session, body_tag, xmpp_elements):
        # increment the request counter
        session.rid  = session.rid + 1
        iq_send = False
        d = None
        
        # send all the elements
        for el in xmpp_elements:
            if el.name == 'iq' and el.firstChildElement() and  el.firstChildElement().name == 'query' \
               and (el.firstChildElement().uri == 'jabber:iq:auth' or \
                    ( el.firstChildElement().hasAttribute('xmlns') and  el.firstChildElement()['xmlns'] == 'jabber:iq:auth')):
                
                # NOTE - temp hack from iq:auth
                #        need to fix this
                # JWChat expects an iq to return

                d = defer.Deferred()
                if use_xmlstream:
                    iq = xmlstream.IQ(session.xmlstream, el['type'])
                else:
                    iq = client.IQ(session.xmlstream, el['type'])
                iq.attributes = el.attributes
                if not iq.hasAttribute('id'):
                    iq.addUniqueId()

                iq.children   = el.children
                if not use_xmlstream:
                    iq.addCallback(self._returnIq, session, d)
                    # need an error back here?
                    iq.send()
                else:
                    iq.send().addCallback(self._cbIq, session, d)
                iq_send = True
                    
            else:
                # something is wrong here, need to figure out what
                # the xmlns will be lost if this is not done
                #punjab.uriCheck(el,NS_BIND)              
                #if el.uri and el.uri != NS_BIND:
                #    el['xmlns'] = el.uri
                # TODO - get rid of this when we stop supporting old versions
                #        of twisted.words
                if el.uri == NS_BIND:
                    el.uri = None
                if el.defaultUri == NS_BIND:
                    el.defaultUri = None
                    
                session.sendRawXml(el)


        if body_tag.hasAttribute('type') and \
           body_tag['type'] == 'terminate':
            d = session.terminate()
        elif not iq_send:
            # normal request
            d = session.poll(d)
            
        return d
        
    def _returnIq(self, cur_session, d, iq):
        """
        A callback from auth iqs
        """
        return cur_session.poll(d)

    def _cbIq(self, iq, cur_session, d):
        """
        A callback from auth iqs
        """
        
        # session.elems.append(iq)
        return cur_session.poll(d)
        
        
    def inSession(self, body):
        """ """
        if body.hasAttribute('sid'):
            if self.sessions.has_key(body['sid']):        
                return True
        return False

    def getXmppElements(self, b, session):
        """
        Get waiting xmpp elements
        """
        for i, obj in enumerate(session.msgs):
            m = session.msgs.pop(0)
            b.addChild(m)            
        for i, obj in enumerate(session.prs):
            p = session.prs.pop(0)
            b.addChild(p)            
        for i, obj in enumerate(session.iqs):
            iq = session.iqs.pop(0)
            b.addChild(iq)
        
        return b

    def endSession(self, cur_session):
        """ end a punjab jabber session """
        d = cur_session.terminate()
        return d
                
# This is the jabberd commponent connect classes and code


class HttpbComponentService(component.Service):
    def componentConnected(self, xmlstream):
        self.xmlstream = xmlstream
        
        
