# Copyright (c) 2005-2006 Christopher Zorn
# See LICENSE.txt for details
# livepunjab.py

import random

from twisted.cred import portal, checkers, credentials, error as credError
from twisted.internet import protocol, reactor, defer
from zope.interface import Interface, implements
from twisted.python import components, log, failure
from twisted.words.protocols.jabber import client, xmlstream, jid
from twisted.words.xish import domish, utility, xpath
from twisted.protocols import basic


from twisted.web import server, resource, static

from zope.interface import implements, Interface
from twisted.python.components import registerAdapter

from nevow import inevow, loaders, livepage, tags, rend, guard, util, url
from nevow.livepage import set, assign, append, js, document, eol


from xmpp.ns import *

import types
import sha

import session

def date():
    import datetime
    d = datetime.datetime.now()
    return str(d)
    # return "[ %2i : %2i : %2i ]" % (d.hour,d.minute,d.second)

"""
XMPP credential classes
"""


class XMPPClient:

    def __init__(self, myJid, xmlstream):
        self.myJid = myJid
        self.xmlstream = xmlstream
        self.roster = None
        self.factory = None

    def addObserver(self, *args, **kwargs):
        self.xmlstream.addObserver(*args, **kwargs)

    def addOnetimeObserver(self, *args, **kwargs):
        self.xmlstream.addOnetimeObserver(*args, **kwargs)
        

    def sendPresence(self, typ = None, status = None, show = None, children = []):
        presence = domish.Element((NS_CLIENT,'presence'))
        if typ:
            presence['type'] = unicode(typ)
            
        for c in children:
            presence.addChild(c)
            

        self.xmlstream.send(presence)

    def getRoster(self, cb=None):
        
        log.msg('Get Roster')
        try:
            iq = client.IQ(self.xmlstream, 'get')
            iq['from'] = self.myJid.full()
            q = iq.addElement('query')
            q['xmlns'] = 'jabber:iq:roster'
            if cb:
                iq.addCallback(cb)

            iq.send()

        except:
            log.err()


    def logout(self):
        log.msg('Logout')
        self.sendPresence(typ='unavaliable')
        self.xmlstream.send("</stream:stream>")
        self.jfactory.stopTrying()
        self.xmlstream = None
        

def rawDataIn(buf):
        log.msg("RECV: %s" % unicode(buf, 'utf-8').encode('ascii', 'replace'))

def rawDataOut(buf):
    log.msg("SEND: %s" % unicode(buf, 'utf-8').encode('ascii', 'replace'))


class XMPPChecker(object):
    implements(checkers.ICredentialsChecker)
    # TODO - other interfaces?
    credentialInterfaces = (credentials.IUsernamePassword,)
    def __init__(self,server=None, port=5222,v=0):
        self.server   = server
        self.port     = int(port)
        self.v        = v
        
        
    def _cbPasswordMatch(self, xs):
        if xs:
            if self.v == 1:
                xs.rawDataInFn = rawDataIn
                xs.rawDataOutFn = rawDataOut
            client = XMPPClient(self.myJid, xs)
            client.jfactory = self.jfactory
            return client
        else:
            return failure.Failure(credError.UnauthorizedLogin())

    def requestAvatarId(self, credentials):
        if credentials.username == "":
            return failure.Failure(credError.UnauthorizedLogin())
        if credentials.password == "":
            return failure.Failure(credError.UnauthorizedLogin())
        
        return defer.maybeDeferred(
            self.login,
            credentials.username,
            credentials.password).addCallback(
            self._cbPasswordMatch)


    def login(self, username, password):
        self.d = defer.Deferred()
        
        self.myJid = jid.JID(username)

        if self.myJid.user is None:
            self.myJid.host = self.server
            self.myJid.user = username
            
        if self.myJid.resource is None:
            self.myJid.resource = 'XMPPCred'

        if self.myJid.host:
            server = self.myJid.host
        else:
            server = self.server

        # parse a port too    
        self.jfactory = client.basicClientFactory(self.myJid, password)
        # TODO - clean this up
        self.jfactory.addBootstrap("//event/stream/authd",self.authd)
        self.jfactory.addBootstrap("//event/client/basicauth/invaliduser", self.authe)
        self.jfactory.addBootstrap("//event/client/basicauth/authfailed", self.authe)
        self.jfactory.addBootstrap("//event/stream/error", self.authe)
        reactor.connectTCP(server,self.port,self.jfactory)
        return self.d
    
    def authd(self, xmlstream):
        if not self.d.called:
            self.d.callback(xmlstream)

    def authe(self, e):
        log.err(e.toXml())
        if not self.d.called:
            self.d.errback(failure.Failure(credError.UnauthorizedLogin()))
        if self.jfactory:
            self.jfactory.stopTrying()
            self.jfactory = None

class IXMPPAvatar(Interface):
    "should have attributes jid, username, host and resource"


class XMPPAvatar:
    implements(IXMPPAvatar)

    def __init__(self, avatar):
        user, xmlstream = avatar
        self.jid      = user
        self.username = self.jid.user
        self.host     = self.jid.host
        self.resource = self.jid.resource
        self.xmlstream = xmlstream
        
### Authentication
def noLogout():
    return None

class XMPPRealm:
    implements(portal.IRealm)

    def requestAvatar(self, avatarId, mind, *interfaces):
        if IXMPPAvatar in interfaces:
            # TODO - pass along xmlstream?
            #        use self.logout when we do
            logout = lambda: None
            return (XMPPAvatar(avatarId),
                    IXMPPAvatar,
                    logout)
        else:
            raise KeyError("None of the requested interfaces is supported") 
            

    def logout(self):
        self.xs.send('</stream:stream>')
        self.xs = None



class PunjabRealm:
    implements(portal.IRealm)

    def requestAvatar(self, avatarId, mind, *interfaces):
        if inevow.IResource in interfaces:
        #if IXMPPAvatar in interfaces:
            # TODO - pass along xmlstream?
            #        use self.logout when we do
            logout = lambda: None
            
            if avatarId is checkers.ANONYMOUS:
                resc = LivePage()
                resc.realm = self
                return (inevow.IResource, resc, noLogout)
            
                # return (XMPPAvatar(avatarId),
                #    IXMPPAvatar,
                #    logout)
            else:
                resc = LivePage(avatarId)
                resc.realm = self
                
                return (inevow.IResource, resc, resc.logout)
 
        else:
            raise KeyError("None of the requested interfaces is supported") 
            

    def logout(self):
        self.xs.send('</stream:stream>')
        self.xs = None

# These tests are examples taken from the twisted book
# http://www.oreilly.com/catalog/twistedadn/
class LoginTestProtocol(basic.LineReceiver):
    def lineReceived(self, line):
        cmd = getattr(self, 'handle_' + self.currentCommand)
        cmd(line.strip())

    def connectionMade(self):
        self.transport.write("User Name: ")
        self.currentCommand = 'user'

    def handle_user(self, username):
        self.username = username
        self.transport.write("Password: ")
        self.currentCommand = 'pass'

    def handle_pass(self, password):
        creds = credentials.UsernamePassword(self.username, password)
        self.factory.portal.login(creds, None, IXMPPAvatar).addCallback(
            self._loginSucceeded).addErrback(
            self._loginFailed)

    def _loginSucceeded(self, avatarInfo):
        avatar, avatarInterface, logout = avatarInfo
        self.transport.write("Welcome %s!\r\n" % avatar.jid.full().encode('utf-8'))
        defer.maybeDeferred(logout).addBoth(self._logoutFinished)

    def _logoutFinished(self, result):
        self.transport.loseConnection()

    def _loginFailed(self, failure):
        self.transport.write("Denied: %s.\r\n" % failure.getErrorMessage())
        self.transport.loseConnection()
    
class LoginTestFactory(protocol.ServerFactory):
    protocol = LoginTestProtocol

    def __init__(self, portal):
        self.portal = portal

        

# resource and nevow classes

class LivePage(livepage.LivePage):
    addSlash = True

    docFactory = loaders.stan(
        tags.html[
        tags.invisible(render=tags.directive("isLogged"))[
         tags.div(pattern="False")[
            tags.head[tags.title['live punjab'],
                ],
            tags.body[
                
                    tags.form(action=guard.LOGIN_AVATAR, method='post')[
                        tags.table[
                            tags.tr[
                                tags.td[ "Username:" ],
                                tags.td[ tags.input(type='text',name='username') ],
                            ],
                            tags.tr[
                                tags.td[ "Password:" ],
                                tags.td[ tags.input(type='password',name='password') ],
                            ]
                        ],
                        tags.input(type='submit'),
                        tags.p,
                    ]]
                
            ],
           tags.invisible(pattern="True")[
              tags.head[tags.title['live punjab'],
                        tags.script(type="text/javascript")["MochiKit = {__compat__: true};"],
                        tags.script(type="text/javascript",
                                    src=url.here.child('MochiKit.js')),
                        livepage.glue,
                        tags.script(type="text/javascript",
                                    src=url.here.child('livepunjab.js')),
                        ],
                        
                tags.body[
                      tags.invisible(render=tags.directive("sessionId")),

                      tags.br,

                      tags.a(href=guard.LOGOUT_AVATAR)["Logout"],
                      

                    ]]
                
            ]])

    
    def __init__(self, avatarId=None):
        rend.Page.__init__(self)
        self.avatarId=avatarId
        self.client = None
        self.messages  = []
        self.presences = []
        self.iqs       = []
        # this will need to be cleaned up
        setattr(self,'child_livepunjab.js',
                    static.File(util.resource_filename('punjab', 'livepunjab.js'),'text/javascript'))
        try:
            setattr(self,'child_MochiKit.js',
                    static.File(util.resource_filename('punjab', 'MochiKit.js'),'text/javascript'))
                        
        except:
            log.err("Error in loading MochiKit, Please download mochikit and place /packed/MochiKit.js into the punjab directory.")
            
    def goingLive(self, ctx, client):
        self.client = client
        for msg in self.messages:
            self.client.send(livepage.alert(str(msg.toXml())))
        for prs in self.presences:
            self.client.send(livepage.alert(str(prs.toXml())))
        for iq in self.iqs:
            self.client.send(livepage.alert(str(iq.toXml())))


    def str2mochi(seld, elem):
        log.msg(str(elem.toStr()))

        mstr = """

        """

        return mstr
        
    def build_roster(self, iq):
        log.msg('Build Roster')
        log.msg(str(iq.toXml()))
        send_roster = """
            roster = IQ{'from':%s};
            alert(roster.emitHTML());
            """ % (iq['from'],)
        
        self.client.send(send_roster)
        #self.client.send(livepage.alert(str(iq.toXml())))
        
    def handle_change(self, ctx, value):
        return livepage.alert(value)


    def client_presence(self, prs):
        log.msg(prs.toXml())
        if self.client:
            self.client.send(livepage.alert(str(prs.toXml())))
        else:
            self.presences.append(prs)
            
    def client_message(self, msg):
        log.msg(msg.toXml())
        if self.client:
            send_msg = """
            m = MESSAGE{'to':%s};
            alert(m.emitHTML());
            """
            self.client.send(send_msg)
        else:
            self.messages.append(msg)
            
    def render_isLogged(self, context, data):
        q = inevow.IQ(context)
        true_pattern = q.onePattern('True')
        false_pattern = q.onePattern('False')
        if self.avatarId:
            self.avatarId.addObserver("/presence", self.client_presence)
            self.avatarId.addObserver("/message", self.client_message)
            #self.avatarId.addObserver("/presence", self.client_presence)
            
            self.avatarId.sendPresence()
            reactor.callLater(1, self.avatarId.getRoster,self.build_roster)
            return true_pattern or context.tag().clear()
        else:
            return false_pattern or context.tag().clear()
 
    def render_sessionId(self, context, data):
        sess = inevow.ISession(context)
        return context.tag[sess.uid]
 
    def logout(self):
        print "Bye"
        if self.avatarId:
            self.avatarId.logout()
            self.avatarId = None

def createResource(**kwargs):
    if kwargs.has_key('server'):
        server = kwargs['server']
    else:
        server = 'thetofu.com'

    if kwargs.has_key('port'):
        port = int(kwargs['port'])
    else:
        port = 5222
        
    realm = PunjabRealm()
    porta = portal.Portal(realm)
    porta.registerChecker(XMPPChecker(server,port))
    porta.registerChecker(checkers.AllowAnonymousAccess(), credentials.IAnonymous)
    
    res = guard.SessionWrapper(porta)
    return res


if __name__ == "__main__":
    p = portal.Portal(XMPPRealm())
    p.registerChecker(XMPPChecker('thetofu.com',5222))
    factory = LoginTestFactory(p)
    reactor.listenTCP(2323, factory)
    reactor.run()


