from twisted.words.protocols import jabber
from twisted.words.protocols.jabber import client,jid

try:
    from twisted.words.xish import domish
except:
    from twisted.xish import domish

from twisted.python import components, failure, log
from twisted.internet import defer, reactor
from twisted.web import server, resource, static
from twisted.cred import portal, checkers, error, credentials


import sha,types

NS_PUBSUB = 'http://jabber.org/protocol/pubsub'
NS_PUBSUB_EVENT = NS_PUBSUB + '#event'
NS_PUBSUB_ERRORS = NS_PUBSUB + '#errors'
NS_PUBSUB_OWNER = NS_PUBSUB + "#owner"

IQ_GET = '/iq[@type="get"]'
IQ_SET = '/iq[@type="set"]'
PUBSUB_ELEMENT = '/pubsub[@xmlns="' + NS_PUBSUB + '"]'
PUBSUB_OWNER_ELEMENT = '/pubsub[@xmlns="' + NS_PUBSUB_OWNER + '"]'
PUBSUB_GET = IQ_GET + PUBSUB_ELEMENT
PUBSUB_SET = IQ_SET + PUBSUB_ELEMENT
PUBSUB_OWNER_GET = IQ_GET + PUBSUB_OWNER_ELEMENT
PUBSUB_OWNER_SET = IQ_SET + PUBSUB_OWNER_ELEMENT
PUBSUB_CREATE = PUBSUB_SET + '/create'
PUBSUB_PUBLISH = PUBSUB_SET + '/publish'
PUBSUB_SUBSCRIBE = PUBSUB_SET + '/subscribe'
PUBSUB_UNSUBSCRIBE = PUBSUB_SET + '/unsubscribe'
PUBSUB_OPTIONS_GET = PUBSUB_GET + '/options'
PUBSUB_OPTIONS_SET = PUBSUB_SET + '/options'
PUBSUB_CONFIGURE_GET = PUBSUB_OWNER_GET + '/configure'
PUBSUB_CONFIGURE_SET = PUBSUB_OWNER_SET + '/configure'
PUBSUB_AFFILIATIONS = PUBSUB_GET + '/affiliations'
PUBSUB_ITEMS = PUBSUB_GET + '/items'
PUBSUB_RETRACT = PUBSUB_SET + '/retract'
PUBSUB_PURGE = PUBSUB_SET + '/purge'
PUBSUB_DELETE = PUBSUB_SET + '/delete'



class PubSub(client.IQ):
    """ A class for pubsub stuff"""
    def __init__(self, xmlstream, to, resource = 1):
         client.IQ.__init__(self,xmlstream,'set')
         self.addElement('pubsub')
         self.pubsub['xmlns'] = NS_PUBSUB
         self['to']   = to
         if resource == 1:
             self['from'] = xmlstream.authenticator.jid.full()
         else:
             self['from'] = xmlstream.authenticator.jid.userhost()
         self.jid     = xmlstream.authenticator.jid
    # TODO - add methods that send too

    def addCreateNode(self, node = None):
        self.children[0].addElement('create')
        self['type'] = 'set'
        if node is not None:
            self.children[0].children[0]['node'] = node
            

    def addPublishNode(self, id, data, itemid = None):
        self['type'] = 'set'
        p = self.children[0].addElement('publish')
        p['node'] = id
        i = p.addElement('item')
        if isinstance(data,str):
            i.rawdata = data
        else:
            i.addChild(data)

    def addConfigureNode(self, id, iqtype):
        self['type'] = iqtype        
        
        #self.uri = NS_PUBSUB_OWNER
        
        self.pubsub['xmlns'] = NS_PUBSUB_OWNER        
        
        p = self.children[0].addElement('configure')
        p['node'] = id

        #if isinstance(data,str):
        #    i.rawdata = data
        #elif data is not None:
        #    i.addChild(data)

    def addDeleteNode(self, id, itemid = 'current'):
        self['type'] = 'set'
        p = self.children[0].addElement('retract')
        p['node'] = id
        i = p.addElement('item')
        i['id'] = itemid
            

    def addSubscribeNode(self, id = None):
        self['type'] = 'set'
        s = self.children[0].addElement('subscribe')
        s['node'] = id
        s['jid']  = self.jid.full()

    def addAffiliationsNode(self):
        self['type'] = 'get'
        s = self.children[0].addElement('affliations')



    def addRetrieveItemsNode(self, id = None, subid = None, max_items = '10'):
        self['type'] = 'get'
        s = self.children[0].addElement('items')
        if id is not None:
            s['node'] = id
        if subid is not None:
            s['subid'] = subid
        s['max_items'] = max_items

